import numpy as np

class Layer():
  def __init__(self,size,activation,lr=0.001,optimizer_name='GD',beta=1,raw=1,epsilon=1 ):
    self.W = np.random.randn(size[1], size[0]) * np.sqrt(1. / size[1])
    self.b = np.zeros(size[1])
    self.activation = activation
    self.lr = lr
    self.optimizer_name=optimizer_name
    self.beta = beta
    self.raw= raw
    self.epsilon = epsilon

  def forward(self,x):
    self.x = x
    self.Z = np.dot(self.W, self.x) + self.b
    self.A = self.activation(self.Z)
    return self.A

  def backward(self,error):
    pass


from softmax_layer import SoftmaxLayer
from evaluation_module import Evaluation_Module
import time
import numpy as np
from FC import FC
from optimizer import optimizer
from activations import relu, sigmoid, softmax
import logging

class DeepNeuralNetwork():
    def __init__(self, epochs=10, l_rate=0.001,optimizer_name='GD',beta=1,raw=1,epsilon=1):
        self.epochs = epochs
        self.l_rate = l_rate
        self.layers = []
        self.optimizer_name = optimizer_name
        self.beta = beta
        self.raw= raw
        self.epsilon = epsilon
       



    def add(self, size,activation):

          self.layers.append(FC(size,activation,lr=self.l_rate,beta=self.beta,raw=self.raw,epsilon=self.epsilon))

    def addout(self, size,activation):

          self.layers.append(SoftmaxLayer(size,activation,lr=self.l_rate,beta=self.beta,raw=self.raw,epsilon=self.epsilon))
        
    def forward_pass(self, x_train):

        # input layer activations becomes sample
        A = x_train


        L = len(self.layers)            # number of layers in the neural network
        
        # Implement [LINEAR -> RELU]*(L-1). Add "cache" to the "caches" list.
        for l in range(L):
            A = self.layers[l].forward(A)

                
        return A


    def backward_pass(self, y_train):
        '''
            This is the backpropagation algorithm, for calculating the updates
            of the neural network's parameters.

            Note: There is a stability issue that causes warnings. This is 
                  caused  by the dot and multiply operations on the huge arrays.
                  
                  RuntimeWarning: invalid value encountered in true_divide
                  RuntimeWarning: overflow encountered in exp
                  RuntimeWarning: overflow encountered in square
        '''
        L = len(self.layers) 
        error = y_train
        for l in range(L-1,0,-1):
        # Calculate  update
          error = self.layers[l].backward(error)
        

        
    def loss(self,A, Y,m):
      cost = -(1.0/m) * np.sum(Y*np.log(A) + (1-Y)*np.log(1-A))
      cost = np.squeeze(cost)
      return cost


    def compute_accuracy(self, x_val, y_val):
        '''
            This function does a forward pass of x, then checks if the indices
            of the maximum value in the output equals the indices in the label
            y. Then it sums over each prediction and calculates the accuracy.
        '''
        predictions = []
        out = np.zeros(y_val.shape)
        labels = []
        i=0
        for x, y in zip(x_val, y_val):
            output = self.forward_pass(x)
            pred = np.argmax(output)
            labels.append(pred)
            out[i][pred] = 1
            i += 1
            predictions.append(pred == np.argmax(y))

        return np.mean(predictions),out, labels

    def train(self, x_train, y_train, x_val, y_val):
        start_time = time.time()
        for iteration in range(self.epochs):
            outputs = []
            for x,y in zip(x_train, y_train):
                output = self.forward_pass(x)
                outputs.append(output)
                self.backward_pass(y)
            
            outputs = np.array(outputs)
            losses = self.loss(outputs,y_train,6000)
            accuracy, predict, labels = self.compute_accuracy(x_val, y_val)
            logging.info('Epoch: {0}, Time Spent: {1:.2f}s, Accuracy: {2:.2f}%, Loss: {3:.4f}'.format(
                iteration+1, time.time() - start_time, accuracy * 100, losses
            ))
        return predict




# -*- coding: utf-8 -*-
"""
Created on Mon Jan 25 00:47:16 2021

@author: reemn
"""

from KhawagaNeuralNetwork.deepNN import DeepNeuralNetwork
from KhawagaNeuralNetwork.evaluation_module import Evaluation_Module
from KhawagaNeuralNetwork.FC import FC
from KhawagaNeuralNetwork.layer import Layer
from KhawagaNeuralNetwork.optimizer import Optimizer
from KhawagaNeuralNetwork.softmax_layer import SoftmaxLayer
 



import numpy as np
from optimizer import optimizer
from layer import Layer

class FC(Layer):
  def __init__(self,size,activation,lr=0.001,optimizer_name='GD',beta=1,raw=1,epsilon=1):
    super(). __init__(size,activation,lr=0.001,optimizer_name='GD',beta=1,raw=1,epsilon=1)

  def backward(self,error):
    
          error = error * self.activation(self.Z, derivative=True)
          change_w = np.outer(error, self.x)
          change_b = np.sum(error,keepdims=True)
          
          w=optimizer(self.W,self.lr,change_w)
          b=optimizer(self.b,self.lr,change_b)

         
          if self.optimizer_name== "momentum_based":
            self.W=w.momentum_based(self.beta)
            self.b=b.momentum_based(self.beta)
          elif self.optimizer_name=="adagrad":
            self.W=w.adagrad()
            self.b=b.adagrad()
          elif self.optimizer_name=="RMSProp":
            self.W=w.RMSProp(self.raw)
            self.b=b.RMSProp(self.raw)
          elif self.optimizer_name=="adaDelta":
            self.W=w.adaDelta(self.raw,self.epsilon)
            self.b=b.adaDelta(self.raw,self.epsilon)
          else: 
            self.W=w.GD()
            self.b=b.GD()
            
          return np.dot(self.W.T, error)



import numpy as np


def sigmoid(x, derivative=False):
        if derivative:
            return (np.exp(-x))/((np.exp(-x)+1)**2)
        return 1/(1 + np.exp(-x))

def softmax( x, derivative=False):
    # Numerically stable with large exponentials
    exps = np.exp(x - x.max())
    if derivative:
        return exps / np.sum(exps, axis=0) * (1 - exps / np.sum(exps, axis=0))
    return exps / np.sum(exps, axis=0)

def relu(x, derivative=False):
    if derivative:
        x[x<=0] = 0
        x[x>0] = 1
        return x
    
    else:
        shapex=x.shape
        k=np.zeros(shapex,dtype=float)
        return np.maximum(k,x)


import numpy as np

class optimizer:

   def __init__(self,weights,lr,delta):
     self.lr=lr
     self.delta=delta
     self.weights=weights
     self.weights=np.array(self.weights)

   def GD (self):
      self.weights=self.weights-np.multiply(self.lr,self.delta)

      return self.weights
    
   
   def momentum_based(self,beta):
      V =np.zeros((self.weights.shape[0],self.weights.shape[1])) 
      V=V+np.multiply(beta,V)-np.multiply(self.lr,self.delta)
      self.weights= self.weights+V
      return self.weights

   
   def Nesstrove(self,beta):
      V =np.zeros((self.weights.shape[0],self.weights.shape[1]))
      

   def adagrad(self):
      A=np.zeros((self.weights.shape[0],self.weights.shape[1]))
      A=np.sum(np.power(self.delta, 2))
      self.weights=self.weights-np.multiply((self.lr/np.sqrt(A)),self.delta)
      return self.weights

   def RMSProp(self,raw):
    
      A=np.zeros((self.weights.shape[0],self.weights.shape[1]))
      A=np.sum(np.multiply(np.multiply(raw,A)+(1-raw),np.power(self.delta,2)))
      self.weights=self.weights-np.multiply((self.lr/np.sqrt(A)),self.delta)
      return self.weights

    
       
 

   def adaDelta(self,raw,epsilon):
      
    A=np.zeros((self.weights.shape[0],self.weights.shape[1]))
    sigma=np.zeros((self.weights.shape[0],self.weights.shape[1]))
    sigma=np.random.rand()
    del_w=np.zeros((self.weights.shape[0],self.weights.shape[1]))
    A=np.sum(np.power(self.delta,2))
    del_w=np.sqrt(np.multiply(sigma/(A+epsilon),self.delta))
    sigma=raw*sigma+(1-raw)*np.power(del_w,2)
    self.weights=self.weights-np.sqrt(sigma/(A+epsilon))*self.delta
    return self.weights 
    
   



   


import pickle

def Save_Model(Model):
  Pkl_Filename = "Pickle_RL_Model.pkl"
  with open(Pkl_Filename, 'wb') as file:  
    pickle.dump(Model, file)
#Loading Model using Pickle, Pass file name
def Load_Model(filename):
  with open(filename, 'rb') as file:  
    Pickled_LR_Model = pickle.load(file)
  return Pickled_LR_Model


import numpy as np
import pandas as pd

class Evaluation_Module:
    def __init__(self, true, pred):
        self.true = np.array(true)
        self.pred = np.array(pred)
        self.Category =range(self.true.shape[1])
        self.lowestCategory = len(self.Category)
        arrayinit = np.zeros((10, 10))
        self.matrix = pd.DataFrame(arrayinit, index=range(self.true.shape[1]), columns=range(self.true.shape[1]))

    def confusion_matrix(self):
        for i in range(self.true.shape[0]):
            for j in range(10):
                if self.true[i][j] == self.pred[i][j] == 1:
                    self.matrix.loc[[self.Category[j]], [self.Category[j]]] += 1

                if self.true[i][j] != self.pred[i][j]:
                    self.matrix.loc[[self.Category[np.argmax(self.pred[i])]], self.Category[np.argmax(self.true[i])]] += 1
                    break


                
        return self.matrix.rename(index = {self.Category[i]: "pred: " + str(i) for i in self.Category}, columns= {self.Category[i]: "true: " + str(i) for i in self.Category})


    def accuracy(self):
        diagonal = np.diag(self.matrix)
        acc = diagonal.sum(dtype=np.float)/(self.matrix.to_numpy().sum(dtype=np.float))
        return acc

    def percision(self):
        arrayinit = np.zeros((1, self.lowestCategory))
        precisionmatrix = pd.DataFrame(arrayinit, index=["percision"], columns=self.Category)
        print(self.lowestCategory)
        for i in range(self.lowestCategory):
            out = (self.matrix.loc[self.Category[i], self.Category[i]])/(self.matrix.loc[self.Category[i],:].to_numpy().sum(dtype=np.float))
            out = np.nan_to_num(out)
            precisionmatrix.loc["percision",self.Category[i]] = out
        return precisionmatrix

    def recall(self):
        arrayinit = np.zeros((1, self.lowestCategory))
        recallmatrix = pd.DataFrame(arrayinit, index=["recall"], columns=self.Category)
        for i in range(self.lowestCategory):
            out = (self.matrix.loc[self.Category[i], self.Category[i]])/(self.matrix.loc[:,self.Category[i]].to_numpy().sum(dtype=np.float))
            out = np.nan_to_num(out)
            recallmatrix.loc["recall",self.Category[i]] = out
        return recallmatrix

    def f1_score(self):
        arrayinit = np.zeros((1, self.lowestCategory))
        f1_scorematrix = pd.DataFrame(arrayinit, index=["f1_score"], columns=self.Category)
        recall = self.recall()
        percision = self.percision()
        for i in range(self.lowestCategory):
            out = 2*(recall.loc["recall",self.Category[i]] * percision.loc["percision",self.Category[i]])/(recall.loc["recall",self.Category[i]] + percision.loc["percision",self.Category[i]])
            out = np.nan_to_num(out)
            f1_scorematrix.loc["f1_score",self.Category[i]] = out
        return f1_scorematrix



import matplotlib.image as mpimg 
import numpy as np
from numpy import savetxt
import os
import pandas as pd 




def load_images_from_folder(folder):
    np.images = []
    np.y=[]
    np.x=[]
    i=0
    for root, dirs, files in os.walk(folder):
        for filename in files:
          np.image = mpimg.imread(folder+ '/' + filename)
          np.image1 = np.image.flatten() 
          np.pixel_size=np.image.shape   
          if np.image is not None:
            np.images.append(np.image1)
            temp = np.zeros(10)
            temp[int(filename[6])] = 1
            np.y.append(temp)
    np.x=np.array(np.y)   
    return np.images, np.x


def normalization(fetaure_matrix):
  x=np.max(np.fetaure_matrix)
  np.y=np.fetaure_matrix/x
  return np.y


    
def load_images_from_folder_compressed(folder):
    np.images = []
    np.y=[]
    i=0
    for filename in os.listdir(folder):
      np.image = mpimg.imread(os.path.join(folder,filename))
      
      np.pixel_size=np.image.shape
      np.fetaure_matrix=np.zeros((np.pixel_size))
      np.dot(np.image[...,:3], [0.2989, 0.5870, 0.1140])
      bottom_half = np.image[np.image.shape[0]//2:, :]
      np.image = np.sum(bottom_half, axis=0)
      
 
  
      if np.image is not None:
        np.images.append(np.image)
    return np.images


def split(feature,label):
  indices = np.random.permutation(feature.shape[0])
  training_idx, test_idx = indices[:int(0.8*feature.shape[0])], indices[int(0.8*feature.shape[0]):]
  traininglabels, testlabels,trainingfeatures,testfeatures = label[training_idx,:],label[test_idx,:],feature[training_idx,:],feature[test_idx,:]
  return traininglabels,testlabels,trainingfeatures,testfeatures

